from __future__ import annotations

from typing import TypeVar, Any, Callable, Awaitable, Type

from pydantic import BaseModel

from xpoz._transform._field_mapping import map_fields_to_camel, map_dict_keys_to_snake
from xpoz._mcp._polling import wait_for_result, wait_for_result_sync
from xpoz._pagination import PaginatedResult, AsyncPaginatedResult
from xpoz.types.common import PaginationInfo

T = TypeVar("T", bound=BaseModel)


def _parse_item(model: Type[T], raw: dict[str, Any]) -> T:
    return model.model_validate(map_dict_keys_to_snake(raw))


def _parse_items(model: Type[T], raw_list: list[dict[str, Any]]) -> list[T]:
    return [_parse_item(model, item) for item in raw_list]


def _extract_pagination(raw: dict[str, Any]) -> PaginationInfo:
    pag: dict[str, Any] = raw.get("pagination") or {}
    return PaginationInfo(
        table_name=pag.get("tableName"),
        total_rows=pag.get("totalRows", 0),
        total_pages=pag.get("totalPages", 0),
        page_number=pag.get("pageNumber", 1),
        page_size=pag.get("pageSize", 100),
        results_count=pag.get("resultsCount", 0),
    )


def _extract_results(raw: dict[str, Any]) -> list[dict[str, Any]]:
    results: list[dict[str, Any]] = raw.get("results", [])
    return results


def _extract_export_op_id(raw: dict[str, Any]) -> str | None:
    val: str | None = raw.get("dataDumpExportOperationId")
    return val


class BaseNamespace:
    def __init__(self, call_tool: Callable[[str, dict[str, Any]], dict[str, Any]], timeout: float):
        self._call_tool = call_tool
        self._timeout = timeout

    def _call_and_maybe_poll(
        self, tool_name: str, arguments: dict[str, Any]
    ) -> dict[str, Any]:
        result = self._call_tool(tool_name, arguments)
        operation_id = result.get("operationId")
        if operation_id:
            return wait_for_result_sync(self._call_tool, operation_id, self._timeout)
        return result

    def _build_paginated_result(
        self,
        raw: dict[str, Any],
        model: Type[T],
        tool_name: str,
        base_args: dict[str, Any],
    ) -> PaginatedResult[T]:
        items = _parse_items(model, _extract_results(raw))
        pagination = _extract_pagination(raw)
        table_name = pagination.table_name
        export_op_id = _extract_export_op_id(raw)

        def fetch_page(page_number: int, tbl: str | None) -> PaginatedResult[T]:
            args = {**base_args, "pageNumber": page_number}
            if tbl:
                args["tableName"] = tbl
            page_raw = self._call_and_maybe_poll(tool_name, args)
            return self._build_paginated_result(page_raw, model, tool_name, base_args)

        def fetch_export(op_id: str) -> str:
            poll_result = wait_for_result_sync(self._call_tool, op_id, self._timeout)
            url: str = poll_result.get("downloadUrl", "")
            return url

        return PaginatedResult(
            data=items,
            pagination=pagination,
            table_name=table_name,
            export_operation_id=export_op_id,
            fetch_page=fetch_page,
            fetch_export=fetch_export,
        )

    def _build_args(self, **kwargs: Any) -> dict[str, Any]:
        args: dict[str, Any] = {}
        for key, value in kwargs.items():
            if value is not None:
                args[key] = value
        return args

    def _convert_fields(self, fields: list[str] | None) -> list[str] | None:
        return map_fields_to_camel(fields)


class AsyncBaseNamespace:
    def __init__(
        self,
        call_tool: Callable[[str, dict[str, Any]], Awaitable[dict[str, Any]]],
        timeout: float,
    ):
        self._call_tool = call_tool
        self._timeout = timeout

    async def _call_and_maybe_poll(
        self, tool_name: str, arguments: dict[str, Any]
    ) -> dict[str, Any]:
        result = await self._call_tool(tool_name, arguments)
        operation_id = result.get("operationId")
        if operation_id:
            return await wait_for_result(self._call_tool, operation_id, self._timeout)
        return result

    async def _build_paginated_result(
        self,
        raw: dict[str, Any],
        model: Type[T],
        tool_name: str,
        base_args: dict[str, Any],
    ) -> AsyncPaginatedResult[T]:
        items = _parse_items(model, _extract_results(raw))
        pagination = _extract_pagination(raw)
        table_name = pagination.table_name
        export_op_id = _extract_export_op_id(raw)

        async def fetch_page(
            page_number: int, tbl: str | None
        ) -> AsyncPaginatedResult[T]:
            args = {**base_args, "pageNumber": page_number}
            if tbl:
                args["tableName"] = tbl
            page_raw = await self._call_and_maybe_poll(tool_name, args)
            return await self._build_paginated_result(page_raw, model, tool_name, base_args)

        async def fetch_export(op_id: str) -> str:
            poll_result = await wait_for_result(self._call_tool, op_id, self._timeout)
            url: str = poll_result.get("downloadUrl", "")
            return url

        return AsyncPaginatedResult(
            data=items,
            pagination=pagination,
            table_name=table_name,
            export_operation_id=export_op_id,
            fetch_page=fetch_page,
            fetch_export=fetch_export,
        )

    def _build_args(self, **kwargs: Any) -> dict[str, Any]:
        args: dict[str, Any] = {}
        for key, value in kwargs.items():
            if value is not None:
                args[key] = value
        return args

    def _convert_fields(self, fields: list[str] | None) -> list[str] | None:
        return map_fields_to_camel(fields)



from __future__ import annotations

from typing import TypeVar, Generic, Callable, Any, Awaitable

from xpoz.types.common import PaginationInfo

T = TypeVar("T")


class PaginatedResult(Generic[T]):
    def __init__(
        self,
        data: list[T],
        pagination: PaginationInfo,
        table_name: str | None,
        export_operation_id: str | None,
        fetch_page: Callable[[int, str | None], PaginatedResult[T]],
        fetch_export: Callable[[str], str] | None,
    ):
        self.data = data
        self.pagination = pagination
        self._table_name = table_name
        self._export_operation_id = export_operation_id
        self._fetch_page = fetch_page
        self._fetch_export = fetch_export

    def has_next_page(self) -> bool:
        return self.pagination.page_number < self.pagination.total_pages

    def next_page(self) -> PaginatedResult[T]:
        if not self.has_next_page():
            raise IndexError("No more pages available")
        return self._fetch_page_result(self.pagination.page_number + 1)

    def get_page(self, page_number: int) -> PaginatedResult[T]:
        if page_number < 1 or page_number > self.pagination.total_pages:
            raise ValueError(
                f"Page {page_number} out of range (1-{self.pagination.total_pages})"
            )
        return self._fetch_page_result(page_number)

    def get_all_pages(self) -> list[T]:
        all_data = list(self.data)
        current_page = self.pagination.page_number
        while current_page < self.pagination.total_pages:
            current_page += 1
            result = self._fetch_page_result(current_page)
            all_data.extend(result.data)
        return all_data

    def export_csv(self) -> str:
        if self._fetch_export is None or self._export_operation_id is None:
            raise RuntimeError("CSV export not available for this result")
        return self._fetch_export(self._export_operation_id)

    def _fetch_page_result(self, page_number: int) -> PaginatedResult[T]:
        return self._fetch_page(page_number, self._table_name)

    def __repr__(self) -> str:
        return (
            f"PaginatedResult(items={len(self.data)}, "
            f"page={self.pagination.page_number}/{self.pagination.total_pages}, "
            f"total={self.pagination.total_rows})"
        )


class AsyncPaginatedResult(Generic[T]):
    def __init__(
        self,
        data: list[T],
        pagination: PaginationInfo,
        table_name: str | None,
        export_operation_id: str | None,
        fetch_page: Callable[[int, str | None], Awaitable[AsyncPaginatedResult[T]]],
        fetch_export: Callable[[str], Awaitable[str]] | None,
    ):
        self.data = data
        self.pagination = pagination
        self._table_name = table_name
        self._export_operation_id = export_operation_id
        self._fetch_page = fetch_page
        self._fetch_export = fetch_export

    def has_next_page(self) -> bool:
        return self.pagination.page_number < self.pagination.total_pages

    async def next_page(self) -> AsyncPaginatedResult[T]:
        if not self.has_next_page():
            raise IndexError("No more pages available")
        return await self._fetch_page_result(self.pagination.page_number + 1)

    async def get_page(self, page_number: int) -> AsyncPaginatedResult[T]:
        if page_number < 1 or page_number > self.pagination.total_pages:
            raise ValueError(
                f"Page {page_number} out of range (1-{self.pagination.total_pages})"
            )
        return await self._fetch_page_result(page_number)

    async def get_all_pages(self) -> list[T]:
        all_data = list(self.data)
        current_page = self.pagination.page_number
        while current_page < self.pagination.total_pages:
            current_page += 1
            result = await self._fetch_page_result(current_page)
            all_data.extend(result.data)
        return all_data

    async def export_csv(self) -> str:
        if self._fetch_export is None or self._export_operation_id is None:
            raise RuntimeError("CSV export not available for this result")
        return await self._fetch_export(self._export_operation_id)

    async def _fetch_page_result(self, page_number: int) -> AsyncPaginatedResult[T]:
        return await self._fetch_page(page_number, self._table_name)

    def __repr__(self) -> str:
        return (
            f"AsyncPaginatedResult(items={len(self.data)}, "
            f"page={self.pagination.page_number}/{self.pagination.total_pages}, "
            f"total={self.pagination.total_rows})"
        )



from __future__ import annotations

import re
from typing import Any


def parse_response_text(text: str) -> dict[str, Any]:
    if not text.strip():
        return {}

    if text.lstrip().startswith("{"):
        import json
        try:
            return json.loads(text)
        except json.JSONDecodeError:
            pass

    lines = text.split("\n")

    if lines[0].startswith("operationId:"):
        return _parse_flat_kv(lines)

    result: dict[str, Any] = {}
    i = 0
    while i < len(lines):
        line = lines[i]

        if not line.strip():
            i += 1
            continue

        array_match = re.match(r"^data\[(\d+)\]\{(.+)\}:", line)
        if array_match:
            fields = [f.strip() for f in array_match.group(2).split(",")]
            i += 1
            rows = _parse_csv_rows(lines, i, indent=2, fields=fields)
            result["results"] = rows
            i += _count_indented(lines, i, 2)
            continue

        yaml_list_match = re.match(r"^data\[(\d+)\]:", line)
        if yaml_list_match:
            i += 1
            rows = _parse_yaml_list(lines, i, indent=2)
            result["results"] = rows
            i += _count_indented(lines, i, 2)
            continue

        if line == "data:":
            i += 1
            data, consumed = _parse_data_block(lines, i)
            result.update(data)
            i += consumed
            continue

        if line.startswith("success:"):
            i += 1
            continue

        kv_match = re.match(r"^(\S+?):\s*(.*)", line)
        if kv_match:
            result[kv_match.group(1)] = _coerce(kv_match.group(2))

        i += 1

    return result


def _parse_flat_kv(lines: list[str]) -> dict[str, Any]:
    result: dict[str, Any] = {}
    for line in lines:
        m = re.match(r"^(\S+?):\s*(.*)", line)
        if m:
            result[m.group(1)] = _coerce(m.group(2))
    return result


def _parse_data_block(lines: list[str], start: int) -> tuple[dict[str, Any], int]:
    return _parse_block(lines, start, indent=2)


def _parse_block(
    lines: list[str], start: int, indent: int
) -> tuple[dict[str, Any], int]:
    result: dict[str, Any] = {}
    prefix = " " * indent
    i = start
    while i < len(lines):
        line = lines[i]

        if not line.startswith(prefix):
            break
        if len(line) > len(prefix) and line[len(prefix)] == " ":
            break

        stripped = line[indent:]

        array_match = re.match(r"^(\w+)\[(\d+)\]\{(.+)\}:", stripped)
        if array_match:
            key = array_match.group(1)
            fields = [f.strip() for f in array_match.group(3).split(",")]
            i += 1
            rows = _parse_csv_rows(lines, i, indent=indent + 2, fields=fields)
            result[key] = rows
            i += _count_indented(lines, i, indent + 2)
            continue

        yaml_list_match = re.match(r"^(\w+)\[(\d+)\]:", stripped)
        if yaml_list_match:
            key = yaml_list_match.group(1)
            i += 1
            rows = _parse_yaml_list(lines, i, indent=indent + 2)
            result[key] = rows
            i += _count_indented(lines, i, indent + 2)
            continue

        kv_match = re.match(r"^(\S+?):\s*(.*)", stripped)
        if kv_match:
            key, val = kv_match.group(1), kv_match.group(2)
            if val:
                result[key] = _coerce(val)
                i += 1
            else:
                child_prefix = " " * (indent + 2)
                if i + 1 < len(lines) and lines[i + 1].startswith(child_prefix):
                    i += 1
                    nested, consumed = _parse_block(lines, i, indent + 2)
                    result[key] = nested
                    i += consumed
                else:
                    result[key] = None
                    i += 1
            continue

        i += 1

    return result, i - start


def _parse_yaml_list(
    lines: list[str], start: int, indent: int
) -> list[dict[str, Any]]:
    rows: list[dict[str, Any]] = []
    prefix = " " * indent
    item_prefix = prefix + "- "
    continuation_prefix = prefix + "  "
    i = start
    while i < len(lines):
        line = lines[i]
        if not line.startswith(prefix):
            break
        if not line.strip():
            i += 1
            continue
        if line.startswith(item_prefix):
            row: dict[str, Any] = {}
            first_kv = line[len(item_prefix):]
            m = re.match(r"^(\S+?):\s*(.*)", first_kv)
            if m:
                row[m.group(1)] = _coerce(m.group(2))
            i += 1
            while i < len(lines):
                cline = lines[i]
                if not cline.startswith(continuation_prefix):
                    break
                if cline.startswith(item_prefix):
                    break
                stripped = cline[len(continuation_prefix):]
                cm = re.match(r"^(\S+?):\s*(.*)", stripped)
                if cm:
                    row[cm.group(1)] = _coerce(cm.group(2))
                i += 1
            rows.append(row)
        else:
            i += 1
    return rows


def _parse_csv_rows(
    lines: list[str], start: int, indent: int, fields: list[str]
) -> list[dict[str, Any]]:
    rows: list[dict[str, Any]] = []
    prefix = " " * indent
    i = start
    while i < len(lines):
        line = lines[i]
        if not line.startswith(prefix):
            break
        raw_row = line[indent:]
        if not raw_row.strip():
            i += 1
            continue
        values = _split_toon_row(raw_row)
        row: dict[str, Any] = {}
        for j, field in enumerate(fields):
            if j < len(values):
                row[field] = _coerce(values[j])
            else:
                row[field] = None
        rows.append(row)
        i += 1
    return rows


def _split_toon_row(line: str) -> list[str]:
    values: list[str] = []
    buf = ""
    in_quotes = False
    i = 0
    while i < len(line):
        ch = line[i]
        if ch == "\\" and in_quotes and i + 1 < len(line):
            buf += ch + line[i + 1]
            i += 2
            continue
        if ch == '"':
            in_quotes = not in_quotes
            buf += ch
            i += 1
            continue
        if ch == "," and not in_quotes:
            values.append(buf.strip())
            buf = ""
            i += 1
            continue
        buf += ch
        i += 1
    if buf or values:
        values.append(buf.strip())
    return values


def _count_indented(lines: list[str], start: int, indent: int) -> int:
    prefix = " " * indent
    count = 0
    i = start
    while i < len(lines):
        if not lines[i].startswith(prefix) and lines[i].strip():
            break
        count += 1
        i += 1
    return count


def _coerce(value: str) -> Any:
    if not value:
        return None

    if value.startswith('"') and value.endswith('"'):
        return value[1:-1]

    if value == "true":
        return True
    if value == "false":
        return False
    if value == "null" or value == "None":
        return None

    try:
        n = int(value)
        if -9007199254740991 <= n <= 9007199254740991:
            return n
        return value
    except ValueError:
        pass
    try:
        return float(value)
    except ValueError:
        pass

    return value



from __future__ import annotations

from pydantic import BaseModel


class InstagramPost(BaseModel, extra="allow"):
    id: str | None = None
    post_type: str | None = None
    user_id: str | None = None
    username: str | None = None
    full_name: str | None = None
    caption: str | None = None
    media_type: str | None = None

    code_url: str | None = None
    image_url: str | None = None
    video_url: str | None = None
    audio_only_url: str | None = None
    profile_pic_url: str | None = None
    video_subtitles_uri: str | None = None

    subtitles: str | None = None
    video_duration: float | None = None

    like_count: int | None = None
    comment_count: int | None = None
    reshare_count: int | None = None
    video_play_count: int | None = None

    location: str | None = None

    created_at: str | None = None
    created_at_timestamp: int | None = None
    created_at_date: str | None = None
    last_fetch: str | None = None
    last_fetch_datetime: str | None = None
    x_last_updated: str | None = None


class InstagramUser(BaseModel, extra="allow"):
    id: str | None = None
    username: str | None = None
    full_name: str | None = None
    biography: str | None = None
    is_private: bool | None = None
    is_verified: bool | None = None

    follower_count: int | None = None
    following_count: int | None = None
    media_count: int | None = None

    profile_pic_url: str | None = None
    profile_pic_id: str | None = None
    profile_url: str | None = None
    external_url: str | None = None
    has_anonymous_profile_picture: bool | None = None

    last_fetch: str | None = None
    last_fetch_datetime: str | None = None
    x_last_updated: str | None = None

    agg_relevance: float | None = None
    relevant_posts_count: int | None = None
    relevant_posts_likes_sum: int | None = None
    relevant_posts_comments_sum: int | None = None
    relevant_posts_reshares_sum: int | None = None
    relevant_posts_video_plays_sum: int | None = None


class InstagramComment(BaseModel, extra="allow"):
    id: str | None = None
    text: str | None = None
    parent_post_id: str | None = None
    parent_post_user_id: str | None = None
    type: str | None = None
    parent_comment_id: str | None = None
    replied_to_comment_id: str | None = None
    child_comment_count: int | None = None

    user_id: str | None = None
    username: str | None = None
    full_name: str | None = None

    like_count: int | None = None
    status: str | None = None
    is_spam: bool | None = None
    has_translation: bool | None = None

    created_at: str | None = None
    created_at_timestamp: int | None = None
    created_at_date: str | None = None
    last_fetch: str | None = None
    last_fetch_datetime: str | None = None
    x_last_updated: str | None = None



class XpozError(Exception):
    pass


class AuthenticationError(XpozError):
    pass


class XpozConnectionError(XpozError):
    pass


class OperationTimeoutError(XpozError):
    def __init__(self, operation_id: str, elapsed_seconds: float):
        self.operation_id = operation_id
        self.elapsed_seconds = elapsed_seconds
        super().__init__(
            f"Operation {operation_id} timed out after {elapsed_seconds:.0f}s"
        )


class OperationFailedError(XpozError):
    def __init__(self, operation_id: str, error: str):
        self.operation_id = operation_id
        self.error = error
        super().__init__(f"Operation {operation_id} failed: {error}")


class OperationCancelledError(XpozError):
    def __init__(self, operation_id: str):
        self.operation_id = operation_id
        super().__init__(f"Operation {operation_id} was cancelled")


class NotFoundError(XpozError):
    pass


class ValidationError(XpozError):
    pass



from __future__ import annotations

from pydantic import BaseModel

from xpoz.types.common import PaginationInfo


class RedditPost(BaseModel, extra="allow"):
    id: str | None = None
    title: str | None = None
    selftext: str | None = None
    url: str | None = None
    permalink: str | None = None
    post_url: str | None = None
    thumbnail: str | None = None

    author_id: str | None = None
    author_username: str | None = None

    subreddit_name: str | None = None
    subreddit_id: str | None = None

    score: int | None = None
    upvotes: int | None = None
    downvotes: int | None = None
    upvote_ratio: float | None = None
    comments_count: int | None = None
    crossposts_count: int | None = None

    is_self: bool | None = None
    is_video: bool | None = None
    is_original_content: bool | None = None
    over18: bool | None = None
    spoiler: bool | None = None
    locked: bool | None = None
    stickied: bool | None = None
    archived: bool | None = None

    link_flair_text: str | None = None
    post_hint: str | None = None
    domain: str | None = None
    crosspost_parent: str | None = None

    created_at: str | None = None
    created_at_timestamp: int | None = None
    created_at_date: str | None = None
    last_fetch: str | None = None
    last_fetch_datetime: str | None = None
    x_last_updated: str | None = None


class RedditUser(BaseModel, extra="allow"):
    id: str | None = None
    username: str | None = None
    profile_url: str | None = None
    profile_pic_url: str | None = None
    snoovatar_img: str | None = None

    link_karma: int | None = None
    comment_karma: int | None = None
    total_karma: int | None = None
    awardee_karma: int | None = None
    awarder_karma: int | None = None

    is_gold: bool | None = None
    is_mod: bool | None = None
    is_employee: bool | None = None
    has_verified_email: bool | None = None
    is_suspended: bool | None = None
    verified: bool | None = None
    is_blocked: bool | None = None
    accept_followers: bool | None = None
    has_subscribed: bool | None = None
    hide_from_robots: bool | None = None
    pref_show_snoovatar: bool | None = None

    profile_description: str | None = None
    profile_banner_url: str | None = None
    profile_title: str | None = None

    created_at: str | None = None
    created_at_timestamp: int | None = None
    created_at_date: str | None = None
    last_fetch: str | None = None
    last_fetch_datetime: str | None = None
    x_last_updated: str | None = None

    agg_relevance: float | None = None
    relevant_posts_count: int | None = None
    relevant_posts_upvotes_sum: int | None = None
    relevant_posts_comments_count_sum: int | None = None


class RedditComment(BaseModel, extra="allow"):
    id: str | None = None
    body: str | None = None
    parent_post_id: str | None = None
    parent_id: str | None = None

    author_id: str | None = None
    author_username: str | None = None

    post_subreddit_name: str | None = None
    post_subreddit_id: str | None = None

    score: int | None = None
    upvotes: int | None = None
    downvotes: int | None = None
    controversiality: int | None = None

    depth: int | None = None
    is_submitter: bool | None = None
    stickied: bool | None = None
    collapsed: bool | None = None
    edited: bool | None = None
    distinguished: str | None = None

    created_at: str | None = None
    created_at_timestamp: int | None = None
    created_at_date: str | None = None
    last_fetch: str | None = None
    last_fetch_datetime: str | None = None
    x_last_updated: str | None = None


class RedditSubreddit(BaseModel, extra="allow"):
    id: str | None = None
    display_name: str | None = None
    title: str | None = None
    public_description: str | None = None
    description: str | None = None

    subscribers_count: int | None = None
    active_user_count: int | None = None

    subreddit_type: str | None = None
    over18: bool | None = None
    lang: str | None = None
    url: str | None = None
    subreddit_url: str | None = None

    icon_img: str | None = None
    banner_img: str | None = None
    header_img: str | None = None
    community_icon: str | None = None

    created_at: str | None = None
    created_at_timestamp: int | None = None
    created_at_date: str | None = None
    last_fetch: str | None = None
    last_fetch_datetime: str | None = None
    x_last_updated: str | None = None

    agg_relevance: float | None = None
    relevant_posts_count: int | None = None
    relevant_posts_upvotes_sum: int | None = None
    relevant_posts_comments_count_sum: int | None = None


class RedditPostWithComments(BaseModel, extra="allow"):
    post: RedditPost | None = None
    comments: list[RedditComment] | None = None
    comments_pagination: PaginationInfo | None = None
    comments_table_name: str | None = None


class SubredditWithPosts(BaseModel, extra="allow"):
    subreddit: RedditSubreddit | None = None
    posts: list[RedditPost] | None = None
    posts_pagination: PaginationInfo | None = None
    posts_table_name: str | None = None



from __future__ import annotations

from typing import Any

from xpoz.namespaces._base import BaseNamespace, AsyncBaseNamespace, _parse_item, _parse_items
from xpoz._pagination import PaginatedResult, AsyncPaginatedResult
from xpoz.types.instagram import InstagramPost, InstagramUser, InstagramComment
from xpoz._config import _tools


class InstagramNamespace(BaseNamespace):
    def get_posts_by_ids(
        self,
        post_ids: list[str],
        *,
        fields: list[str] | None = None,
        force_latest: bool | None = None,
    ) -> list[InstagramPost]:
        args = self._build_args(
            postIds=post_ids,
            fields=self._convert_fields(fields),
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.GET_INSTAGRAM_POSTS_BY_IDS, args)
        return _parse_items(InstagramPost, result.get("results", []))

    def get_posts_by_user(
        self,
        identifier: str,
        identifier_type: str = "username",
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        force_latest: bool | None = None,
    ) -> PaginatedResult[InstagramPost]:
        args = self._build_args(
            identifier=identifier,
            identifierType=identifier_type,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.GET_INSTAGRAM_POSTS_BY_USER, args)
        return self._build_paginated_result(
            result, InstagramPost, _tools.GET_INSTAGRAM_POSTS_BY_USER, args
        )

    def search_posts(
        self,
        query: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        force_latest: bool | None = None,
    ) -> PaginatedResult[InstagramPost]:
        args = self._build_args(
            query=query,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.SEARCH_INSTAGRAM_POSTS, args)
        return self._build_paginated_result(
            result, InstagramPost, _tools.SEARCH_INSTAGRAM_POSTS, args
        )

    def get_comments(
        self,
        post_id: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        force_latest: bool | None = None,
    ) -> PaginatedResult[InstagramComment]:
        args = self._build_args(
            postId=post_id,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.GET_INSTAGRAM_COMMENTS, args)
        return self._build_paginated_result(
            result, InstagramComment, _tools.GET_INSTAGRAM_COMMENTS, args
        )

    def get_user(
        self,
        identifier: str,
        identifier_type: str = "username",
        *,
        fields: list[str] | None = None,
    ) -> InstagramUser:
        args = self._build_args(
            identifier=identifier,
            identifierType=identifier_type,
            fields=self._convert_fields(fields),
        )
        result = self._call_and_maybe_poll(_tools.GET_INSTAGRAM_USER, args)
        results = result.get("results", [])
        if isinstance(results, list) and len(results) > 0:
            return _parse_item(InstagramUser, results[0])
        return _parse_item(InstagramUser, result)

    def search_users(
        self,
        name: str,
        *,
        limit: int | None = None,
        fields: list[str] | None = None,
    ) -> list[InstagramUser]:
        args = self._build_args(
            name=name,
            limit=limit,
            fields=self._convert_fields(fields),
        )
        result = self._call_and_maybe_poll(_tools.SEARCH_INSTAGRAM_USERS, args)
        return _parse_items(InstagramUser, result.get("results", []))

    def get_user_connections(
        self,
        username: str,
        connection_type: str,
        *,
        fields: list[str] | None = None,
        force_latest: bool | None = None,
    ) -> PaginatedResult[InstagramUser]:
        args = self._build_args(
            username=username,
            connectionType=connection_type,
            fields=self._convert_fields(fields),
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.GET_INSTAGRAM_USER_CONNECTIONS, args)
        return self._build_paginated_result(
            result, InstagramUser, _tools.GET_INSTAGRAM_USER_CONNECTIONS, args
        )

    def get_post_interacting_users(
        self,
        post_id: str,
        interaction_type: str,
        *,
        fields: list[str] | None = None,
        force_latest: bool | None = None,
    ) -> PaginatedResult[InstagramUser]:
        args = self._build_args(
            postId=post_id,
            interactionType=interaction_type,
            fields=self._convert_fields(fields),
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.GET_INSTAGRAM_POST_INTERACTING_USERS, args)
        return self._build_paginated_result(
            result, InstagramUser, _tools.GET_INSTAGRAM_POST_INTERACTING_USERS, args
        )

    def get_users_by_keywords(
        self,
        query: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        force_latest: bool | None = None,
    ) -> PaginatedResult[InstagramUser]:
        args = self._build_args(
            query=query,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.GET_INSTAGRAM_USERS_BY_KEYWORDS, args)
        return self._build_paginated_result(
            result, InstagramUser, _tools.GET_INSTAGRAM_USERS_BY_KEYWORDS, args
        )


class AsyncInstagramNamespace(AsyncBaseNamespace):
    async def get_posts_by_ids(
        self,
        post_ids: list[str],
        *,
        fields: list[str] | None = None,
        force_latest: bool | None = None,
    ) -> list[InstagramPost]:
        args = self._build_args(
            postIds=post_ids,
            fields=self._convert_fields(fields),
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.GET_INSTAGRAM_POSTS_BY_IDS, args)
        return _parse_items(InstagramPost, result.get("results", []))

    async def get_posts_by_user(
        self,
        identifier: str,
        identifier_type: str = "username",
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        force_latest: bool | None = None,
    ) -> AsyncPaginatedResult[InstagramPost]:
        args = self._build_args(
            identifier=identifier,
            identifierType=identifier_type,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.GET_INSTAGRAM_POSTS_BY_USER, args)
        return await self._build_paginated_result(
            result, InstagramPost, _tools.GET_INSTAGRAM_POSTS_BY_USER, args
        )

    async def search_posts(
        self,
        query: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        force_latest: bool | None = None,
    ) -> AsyncPaginatedResult[InstagramPost]:
        args = self._build_args(
            query=query,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.SEARCH_INSTAGRAM_POSTS, args)
        return await self._build_paginated_result(
            result, InstagramPost, _tools.SEARCH_INSTAGRAM_POSTS, args
        )

    async def get_comments(
        self,
        post_id: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        force_latest: bool | None = None,
    ) -> AsyncPaginatedResult[InstagramComment]:
        args = self._build_args(
            postId=post_id,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.GET_INSTAGRAM_COMMENTS, args)
        return await self._build_paginated_result(
            result, InstagramComment, _tools.GET_INSTAGRAM_COMMENTS, args
        )

    async def get_user(
        self,
        identifier: str,
        identifier_type: str = "username",
        *,
        fields: list[str] | None = None,
    ) -> InstagramUser:
        args = self._build_args(
            identifier=identifier,
            identifierType=identifier_type,
            fields=self._convert_fields(fields),
        )
        result = await self._call_and_maybe_poll(_tools.GET_INSTAGRAM_USER, args)
        results = result.get("results", [])
        if isinstance(results, list) and len(results) > 0:
            return _parse_item(InstagramUser, results[0])
        return _parse_item(InstagramUser, result)

    async def search_users(
        self,
        name: str,
        *,
        limit: int | None = None,
        fields: list[str] | None = None,
    ) -> list[InstagramUser]:
        args = self._build_args(
            name=name,
            limit=limit,
            fields=self._convert_fields(fields),
        )
        result = await self._call_and_maybe_poll(_tools.SEARCH_INSTAGRAM_USERS, args)
        return _parse_items(InstagramUser, result.get("results", []))

    async def get_user_connections(
        self,
        username: str,
        connection_type: str,
        *,
        fields: list[str] | None = None,
        force_latest: bool | None = None,
    ) -> AsyncPaginatedResult[InstagramUser]:
        args = self._build_args(
            username=username,
            connectionType=connection_type,
            fields=self._convert_fields(fields),
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.GET_INSTAGRAM_USER_CONNECTIONS, args)
        return await self._build_paginated_result(
            result, InstagramUser, _tools.GET_INSTAGRAM_USER_CONNECTIONS, args
        )

    async def get_post_interacting_users(
        self,
        post_id: str,
        interaction_type: str,
        *,
        fields: list[str] | None = None,
        force_latest: bool | None = None,
    ) -> AsyncPaginatedResult[InstagramUser]:
        args = self._build_args(
            postId=post_id,
            interactionType=interaction_type,
            fields=self._convert_fields(fields),
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.GET_INSTAGRAM_POST_INTERACTING_USERS, args)
        return await self._build_paginated_result(
            result, InstagramUser, _tools.GET_INSTAGRAM_POST_INTERACTING_USERS, args
        )

    async def get_users_by_keywords(
        self,
        query: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        force_latest: bool | None = None,
    ) -> AsyncPaginatedResult[InstagramUser]:
        args = self._build_args(
            query=query,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.GET_INSTAGRAM_USERS_BY_KEYWORDS, args)
        return await self._build_paginated_result(
            result, InstagramUser, _tools.GET_INSTAGRAM_USERS_BY_KEYWORDS, args
        )



from __future__ import annotations

import time
from typing import Any, Callable, Awaitable

import anyio

from xpoz._exceptions import (
    OperationTimeoutError,
    OperationFailedError,
    OperationCancelledError,
)

POLL_INTERVAL_SECONDS = 5
DEFAULT_TIMEOUT_SECONDS = 300


async def wait_for_result(
    call_tool: Callable[[str, dict[str, Any]], Awaitable[dict[str, Any]]],
    operation_id: str,
    timeout: float = DEFAULT_TIMEOUT_SECONDS,
) -> dict[str, Any]:
    start = anyio.current_time()
    while True:
        result: dict[str, Any] = await call_tool(
            "checkOperationStatus", {"operationId": operation_id}
        )
        status = result.get("status")

        if status == "failed":
            error = result.get("error", "Unknown error")
            raise OperationFailedError(operation_id, str(error))

        if status == "cancelled":
            raise OperationCancelledError(operation_id)

        if status == "completed" or "results" in result or "downloadUrl" in result:
            return result

        elapsed = anyio.current_time() - start
        if elapsed >= timeout:
            raise OperationTimeoutError(operation_id, elapsed)

        await anyio.sleep(POLL_INTERVAL_SECONDS)


def wait_for_result_sync(
    call_tool: Callable[[str, dict[str, Any]], dict[str, Any]],
    operation_id: str,
    timeout: float = DEFAULT_TIMEOUT_SECONDS,
) -> dict[str, Any]:
    start = time.monotonic()
    while True:
        result: dict[str, Any] = call_tool(
            "checkOperationStatus", {"operationId": operation_id}
        )
        status = result.get("status")

        if status == "failed":
            error = result.get("error", "Unknown error")
            raise OperationFailedError(operation_id, str(error))

        if status == "cancelled":
            raise OperationCancelledError(operation_id)

        if status == "completed" or "results" in result or "downloadUrl" in result:
            return result

        elapsed = time.monotonic() - start
        if elapsed >= timeout:
            raise OperationTimeoutError(operation_id, elapsed)

        time.sleep(POLL_INTERVAL_SECONDS)



from __future__ import annotations

GET_TWITTER_POSTS_BY_IDS = "getTwitterPostsByIds"
GET_TWITTER_POSTS_BY_AUTHOR = "getTwitterPostsByAuthor"
SEARCH_TWITTER_POSTS = "getTwitterPostsByKeywords"
GET_TWITTER_RETWEETS = "getTwitterPostRetweets"
GET_TWITTER_QUOTES = "getTwitterPostQuotes"
GET_TWITTER_COMMENTS = "getTwitterPostComments"
GET_TWITTER_POST_INTERACTING_USERS = "getTwitterPostInteractingUsers"
COUNT_TWEETS = "countTweets"
GET_TWITTER_USER = "getTwitterUser"
SEARCH_TWITTER_USERS = "searchTwitterUsers"
GET_TWITTER_USER_CONNECTIONS = "getTwitterUserConnections"
GET_TWITTER_USERS_BY_KEYWORDS = "getTwitterUsersByKeywords"

GET_INSTAGRAM_POSTS_BY_IDS = "getInstagramPostsByIds"
GET_INSTAGRAM_POSTS_BY_USER = "getInstagramPostsByUser"
SEARCH_INSTAGRAM_POSTS = "getInstagramPostsByKeywords"
GET_INSTAGRAM_COMMENTS = "getInstagramCommentsByPostId"
GET_INSTAGRAM_USER = "getInstagramUser"
SEARCH_INSTAGRAM_USERS = "searchInstagramUsers"
GET_INSTAGRAM_USER_CONNECTIONS = "getInstagramUserConnections"
GET_INSTAGRAM_POST_INTERACTING_USERS = "getInstagramPostInteractingUsers"
GET_INSTAGRAM_USERS_BY_KEYWORDS = "getInstagramUsersByKeywords"

SEARCH_REDDIT_POSTS = "getRedditPostsByKeywords"
GET_REDDIT_POST_WITH_COMMENTS = "getRedditPostWithCommentsById"
SEARCH_REDDIT_COMMENTS = "getRedditCommentsByKeywords"
GET_REDDIT_USER = "getRedditUser"
SEARCH_REDDIT_USERS = "searchRedditUsers"
GET_REDDIT_USERS_BY_KEYWORDS = "getRedditUsersByKeywords"
SEARCH_REDDIT_SUBREDDITS = "searchRedditSubreddits"
GET_REDDIT_SUBREDDIT_WITH_POSTS = "getRedditSubredditWithPostsByName"
GET_REDDIT_SUBREDDITS_BY_KEYWORDS = "getRedditSubredditsByKeywords"



import re

_CAMEL_TO_SNAKE_RE1 = re.compile(r"(.)([A-Z][a-z]+)")
_CAMEL_TO_SNAKE_RE2 = re.compile(r"([a-z0-9])([A-Z])")


def camel_to_snake(name: str) -> str:
    s = _CAMEL_TO_SNAKE_RE1.sub(r"\1_\2", name)
    return _CAMEL_TO_SNAKE_RE2.sub(r"\1_\2", s).lower()


def snake_to_camel(name: str) -> str:
    parts = name.split("_")
    return parts[0] + "".join(p.capitalize() for p in parts[1:])


def map_fields_to_camel(fields: list[str] | None) -> list[str] | None:
    if fields is None:
        return None
    return [snake_to_camel(f) for f in fields]


def map_dict_keys_to_snake(data: dict[str, object]) -> dict[str, object]:
    return {camel_to_snake(k): v for k, v in data.items()}


def map_list_of_dicts_to_snake(data: list[dict[str, object]]) -> list[dict[str, object]]:
    return [map_dict_keys_to_snake(item) for item in data]



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, field_validator


class TwitterPost(BaseModel, extra="allow"):
    id: str | None = None
    text: str | None = None
    author_id: str | None = None
    author_username: str | None = None
    conversation_id: str | None = None
    lang: str | None = None
    source: str | None = None
    status: str | None = None
    deleted: bool | None = None
    suspended: bool | None = None
    possibly_sensitive: bool | None = None
    is_retweet: bool | None = None

    like_count: int | None = None
    retweet_count: int | None = None
    reply_count: int | None = None
    quote_count: int | None = None
    impression_count: int | None = None
    bookmark_count: int | None = None

    quoted_tweet_id: str | None = None
    retweeted_tweet_id: str | None = None
    reply_to_tweet_id: str | None = None
    reply_to_user_id: str | None = None
    reply_to_username: str | None = None
    original_tweet_id: str | None = None
    edited_tweets: list[str] | None = None
    reply_settings: str | None = None

    hashtags: list[str] | None = None
    mentions: list[str] | None = None
    media_urls: list[str] | None = None
    grok_generated_content: list[dict[str, Any]] | None = None
    urls: list[str] | None = None

    country: str | None = None
    region: str | None = None
    city: str | None = None

    has_birdwatch_notes: bool | None = None
    birdwatch_notes_id: str | None = None
    birdwatch_notes_text: str | None = None
    birdwatch_notes_url: str | None = None

    created_at: str | None = None
    created_at_date: str | None = None
    x_fetched_at: str | None = None


class TwitterUser(BaseModel, extra="allow"):
    id: str | None = None
    username: str | None = None

    @field_validator("username", mode="before")
    @classmethod
    def coerce_username(cls, v: Any) -> Any:
        if v is not None and not isinstance(v, str):
            return str(v)
        return v
    name: str | None = None
    description: str | None = None
    location: str | None = None
    verified: bool | None = None
    verified_type: str | None = None
    protected: bool | None = None
    status: str | None = None

    followers_count: int | None = None
    following_count: int | None = None
    tweet_count: int | None = None
    listed_count: int | None = None
    likes_count: int | None = None
    media_count: int | None = None

    profile_image_url: str | None = None
    profile_banner_url: str | None = None
    profile_interstitial_type: str | None = None

    pinned_tweet_id: str | None = None
    source: str | None = None
    is_verified: bool | None = None
    account_based_in: str | None = None
    location_accurate: bool | None = None
    label: str | None = None
    label_type: str | None = None

    collected_following_count: int | None = None
    collected_followers_count: int | None = None
    collected_followers_coverage: float | None = None
    collected_following_coverage: float | None = None
    avg_tweets_per_day_last_month: float | None = None

    n_lang: int | None = None
    n_langs_filtered: int | None = None
    inauthentic_type: str | None = None
    is_inauthentic: bool | None = None
    is_inauthentic_prob_score: float | None = None
    is_inauthentic_calculated_at: str | None = None

    verified_since_datetime: str | None = None
    username_changes: list[str] | None = None
    last_username_change_datetime: str | None = None

    created_at: str | None = None
    created_at_date: str | None = None
    x_fetched_at: str | None = None
    modified_at: str | None = None
    x_modified_at: str | None = None

    agg_relevance: float | None = None
    relevant_tweets_count: int | None = None
    relevant_tweets_impressions_sum: int | None = None
    relevant_tweets_likes_sum: int | None = None
    relevant_tweets_quotes_sum: int | None = None
    relevant_tweets_replies_sum: int | None = None
    relevant_tweets_retweets_sum: int | None = None



import pytest

from xpoz import PaginatedResult
from xpoz.types.twitter import TwitterPost, TwitterUser
from xpoz.types.common import PaginationInfo


@pytest.fixture(scope="module")
def twitter_search_result(client):
    return client.twitter.search_posts(
        "bitcoin",
        start_date="2025-01-01",
        fields=["id", "text", "like_count", "retweet_count"],
    )


@pytest.fixture(scope="module")
def twitter_paging_result(client):
    return client.twitter.search_posts(
        "bitcoin",
        start_date="2025-01-01",
        fields=["id", "text", "like_count", "retweet_count"],
        response_type="paging",
    )


@pytest.fixture(scope="module")
def twitter_csv_result(client):
    return client.twitter.search_posts(
        "bitcoin",
        start_date="2025-01-01",
        response_type="csv",
    )


@pytest.fixture(scope="module")
def twitter_post_id():
    return "1874266108200673750"


class TestTwitterUsers:
    def test_get_user(self, client):
        user = client.twitter.get_user("elonmusk")
        assert isinstance(user, TwitterUser)
        assert user.username == "elonmusk"

    def test_get_user_by_id(self, client):
        user = client.twitter.get_user("44196397", identifier_type="id")
        assert isinstance(user, TwitterUser)
        assert user.id == "44196397"

    def test_get_user_with_fields(self, client):
        user = client.twitter.get_user(
            "elonmusk", fields=["id", "username", "followers_count"]
        )
        assert isinstance(user, TwitterUser)
        assert user.id is not None
        assert user.username is not None
        assert user.followers_count is not None

    def test_search_users(self, client):
        users = client.twitter.search_users("elon")
        assert isinstance(users, list)
        assert len(users) > 0
        for u in users:
            assert isinstance(u, TwitterUser)
            assert u.username is not None

    def test_get_user_connections(self, client):
        result = client.twitter.get_user_connections("elonmusk", "followers")
        assert isinstance(result, PaginatedResult)
        assert isinstance(result.pagination, PaginationInfo)
        assert result.pagination.total_rows > 0
        assert len(result.data) > 0
        for u in result.data:
            assert isinstance(u, TwitterUser)

    def test_get_users_by_keywords(self, client):
        result = client.twitter.get_users_by_keywords("artificial intelligence")
        assert isinstance(result, PaginatedResult)
        assert len(result.data) > 0
        for u in result.data:
            assert isinstance(u, TwitterUser)


class TestTwitterPosts:
    def test_get_posts_by_author(self, client):
        result = client.twitter.get_posts_by_author(
            "elonmusk", fields=["id", "text", "like_count"]
        )
        assert isinstance(result, PaginatedResult)
        assert len(result.data) > 0
        for post in result.data:
            assert isinstance(post, TwitterPost)
            assert post.text is not None

    def test_search_posts(self, twitter_search_result):
        result = twitter_search_result
        assert isinstance(result, PaginatedResult)
        assert isinstance(result.pagination, PaginationInfo)
        assert result.pagination.page_number == 1
        assert isinstance(result.pagination.total_pages, int)

    def test_search_posts_pagination(self, twitter_paging_result):
        if not twitter_paging_result.has_next_page():
            pytest.skip("Only one page of results")
        page2 = twitter_paging_result.next_page()
        assert isinstance(page2, PaginatedResult)
        assert len(page2.data) > 0

    def test_get_retweets(self, client, twitter_post_id):
        result = client.twitter.get_retweets(twitter_post_id)
        assert isinstance(result, PaginatedResult)

    def test_get_quotes(self, client, twitter_post_id):
        result = client.twitter.get_quotes(twitter_post_id)
        assert isinstance(result, PaginatedResult)

    def test_get_comments(self, client, twitter_post_id):
        result = client.twitter.get_comments(twitter_post_id)
        assert isinstance(result, PaginatedResult)

    def test_get_post_interacting_users(self, client, twitter_post_id):
        result = client.twitter.get_post_interacting_users(
            twitter_post_id, "commenters"
        )
        assert isinstance(result, PaginatedResult)

    def test_get_posts_by_ids(self, client, twitter_post_id):
        posts = client.twitter.get_posts_by_ids([twitter_post_id])
        assert isinstance(posts, list)
        assert len(posts) == 1
        assert isinstance(posts[0], TwitterPost)

    def test_count_posts(self, client):
        count = client.twitter.count_posts("bitcoin", start_date="2025-01-01")
        assert isinstance(count, int)
        assert count > 0

    def test_export_csv(self, twitter_csv_result):
        url = twitter_csv_result.export_csv()
        assert isinstance(url, str)
        assert len(url) > 0
